@@ -83,6 +83,33 @@ HTML_ACCESSIBLE_FACTORY (HTML_TYPE_BOX_ACCESSIBLE, html_box_accessible)
 HTML_ACCESSIBLE_FACTORY (HTML_TYPE_BOX_TABLE_ACCESSIBLE, html_box_table_accessible)
 #endif
 
+static GQuark quark_moving_focus_out = 0;
+
+static gboolean
+is_moving_focus_out (HtmlView *view)
+{
+	gint val; 
+
+	if (quark_moving_focus_out == 0)
+		return FALSE;
+
+	val = GPOINTER_TO_INT (g_object_get_qdata (G_OBJECT (view),
+						   quark_moving_focus_out));
+	return (val != 0); 
+}
+
+static void
+set_moving_focus_out (HtmlView *view, gboolean ret)
+{
+	gint val;
+
+	if (quark_moving_focus_out == 0)
+		quark_moving_focus_out = g_quark_from_static_string ("html-view-moving-focus-out");
+	val = ret ? 1 : 0;
+	g_object_set_qdata (G_OBJECT (view), 
+			    quark_moving_focus_out,
+			    GINT_TO_POINTER (val));
+}
 
 static gboolean
 set_adjustment_clamped (GtkAdjustment *adj, gdouble val)
@@ -211,6 +238,24 @@ html_view_relayout (HtmlView *view)
 
 		if (view->jump_to_anchor)
 			html_view_try_jump (view);
+#ifdef ENABLE_ACCESSIBILITY
+		{
+		AtkObject *child;
+		/*
+		 * If a new HTML page has been displayed and accessibility 
+		 * is enabled we wish to notify that relayout has been called.
+		 * The accessibility code will be able to figure out if a new 
+		 * root has been added to the HtmlView.
+		 *
+		 * The following code achieves that purpose. It would be
+		 * better to define a new signal on HtmlView but that would
+		 * break binary compatibility.
+		 */ 
+		child  = atk_object_ref_accessible_child (gtk_widget_get_accessible (GTK_WIDGET (view)), 0); 
+		if (child)
+			g_object_unref (child);
+		}
+#endif
 	}
 	if (view->relayout_timeout_id != 0) {
 		g_source_remove (view->relayout_timeout_id);
@@ -222,7 +267,8 @@ html_view_relayout (HtmlView *view)
 	}
 
 	if (GTK_WIDGET_HAS_FOCUS (view)) {
-		if (view->document->focus_element == NULL) {
+		if (view->document->focus_element == NULL &&
+		    view->document->dom_document) {
 			html_document_update_focus_element (view->document, html_focus_iterator_next_element (view->document->dom_document, NULL));
 			html_view_focus_element (view);
 		}
@@ -532,16 +578,31 @@ html_view_repaint_callback (HtmlDocument *document, DomNode *node, HtmlView *vie
 static void
 html_view_style_updated (HtmlDocument *document, DomNode *node, HtmlStyleChange style_change, HtmlView *view)
 {
-	HtmlBox *box = html_view_find_layout_box (view, node, FALSE);
+	HtmlBox *box;
 	DomNode *child_node;
+	HtmlStyle *style;
 
+	if (node == NULL)
+		return;
+
+	box = html_view_find_layout_box (view, node, FALSE);
 	for (child_node = dom_Node__get_firstChild (node); child_node; child_node = dom_Node__get_nextSibling (child_node)) {
 		html_view_style_updated (document, child_node, style_change, view);
 	}
-
 	if (!box)
 		return;
 
+	style = HTML_BOX_GET_STYLE (box);
+	if (DOM_IS_ELEMENT (node) && 
+		dom_element_is_focusable (DOM_ELEMENT (node))) {
+		gint focus_width;
+
+		gtk_widget_style_get (GTK_WIDGET (view),
+				      "focus-line-width", &focus_width,
+				      NULL);
+		html_style_set_outline_width (style, 2 * focus_width);			
+	}	
+
 	switch (style_change) {
 	case HTML_STYLE_CHANGE_REPAINT:
 		html_view_repaint_callback (document, node, view);
@@ -861,6 +922,14 @@ html_view_focus (GtkWidget *widget, GtkDirectionType direction)
         if (view->document == NULL || view->document->dom_document == NULL) {
 		return FALSE;
 	}
+        if (is_moving_focus_out (view)) {
+                /*
+                 * Widget will retain focus if there is nothing else
+                 * to get focus
+                 */
+                set_moving_focus_out (view, FALSE);
+                return FALSE;
+        }
 
 	if (direction == GTK_DIR_TAB_FORWARD) {
 		new_focus_element = html_focus_iterator_next_element (view->document->dom_document, view->document->focus_element);
@@ -951,11 +1020,16 @@ html_view_real_activate (HtmlView *view)
 static void html_view_real_move_focus_out (HtmlView *view, GtkDirectionType direction_type)
 {
 	GtkWidget *widget = GTK_WIDGET (view);
+	GtkWidget *toplevel;
 
 	html_document_update_focus_element (view->document, NULL);
-	if (!GTK_WIDGET_HAS_FOCUS (widget))
-		gtk_widget_grab_focus (widget);
-	gtk_widget_queue_draw (widget);
+
+	set_moving_focus_out (view, TRUE);
+	toplevel = gtk_widget_get_toplevel (widget);
+	g_return_if_fail (toplevel);
+
+	gtk_widget_child_focus (toplevel, direction_type);
+	set_moving_focus_out (view, FALSE);
 }
 
 static void
@@ -1280,7 +1354,7 @@ html_view_get_type (void)
 }
 
 static void 
-html_view_update_box_style_size (HtmlBox *root, gfloat adjust, GPtrArray *done)
+html_view_update_box_style_size (HtmlBox *root, gfloat adjust, gint focus_width, GPtrArray *done)
 {
 	HtmlBox *box;
 
@@ -1307,9 +1381,14 @@ html_view_update_box_style_size (HtmlBox *root, gfloat adjust, GPtrArray *done)
 					font_spec->size = font_spec->size * adjust;
 				}
 			}
+			if (DOM_IS_ELEMENT (root->dom_node) && 
+			    dom_element_is_focusable (DOM_ELEMENT (root->dom_node))) {
+				html_style_set_outline_width (style, 2 * focus_width);			
+			}	
 		}
+		html_box_set_unrelayouted_up (box);		
 		if (box->children) {
-			html_view_update_box_style_size (box->children, adjust, done);
+			html_view_update_box_style_size (box->children, adjust, focus_width, done);
 		}
 	}
 }
@@ -1321,6 +1400,12 @@ html_view_style_set (GtkWidget *widget, GtkStyle *previous_style)
 		gfloat fsize;
 		gint new_isize;
 		gint old_isize;
+		gint focus_width;
+
+		gtk_widget_style_get (widget,
+				      "focus-line-width", &focus_width,
+				      NULL);
+
 
  		widget->style->bg[GTK_STATE_NORMAL] =
  			widget->style->base[GTK_STATE_NORMAL];
@@ -1358,7 +1443,7 @@ html_view_style_set (GtkWidget *widget, GtkStyle *previous_style)
 					old_size = old_isize;
 					new_size = new_isize;
 				}
-				html_view_update_box_style_size (view->root, (gfloat) new_isize / (gfloat) old_isize, done);
+				html_view_update_box_style_size (view->root, (gfloat) new_isize / (gfloat) old_isize, focus_width, done);
 			}
 		}
 	}
